#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2019 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

"""
test_bgp_large_community_topo_1.py: Test BGP large community.

Following tests are covered:
1. Verify the standard large-community-lists can permit or deny
   large community attribute only in the correct canonical format.
2. Verify the expanded large-community-lists can permit or deny
   large community attribute both in the correct canonical format
   as well as REG_EX.
3. Verify that we can modify a large-community-list is in use,
   to add/remove attribute value and it takes immediate effect.
4. Verify that large community attribute gets advertised when
   route-map is applied to a neighbor and cleared when route-map
   is removed.
5. Verify that duplicate BGP Large Community values are NOT be transmitted.
6. Verify if we want to remove all the large-community attributes from a
   set of prefix we can set the value as NONE.
7. Redistribute connected and static routes in BGP process with a route-map
   appending/removing L-comm attributes.
8. Verify if we want to remove specific large-community values from
   a set of prefix we can make use of DELETE operation based on L-comm list.
9. Verify that if community values are NOT be advertised to a specific
   neighbour, we negate send-community command.
   (Send-community all is enabled by default for all neighbors)
10. Verify that large-community lists can not be configured without providing
    specific L-community values(for match/delete operation in a route-map).
11. Verify that Match_EXACT clause should pass only if all of the L-comm
    values configured (horizontally) in the community list is present in
    the prefix.  There must be no additional L-communities in the prefix.
12. Verify that Match_ALL clause should pass only if ALL of the L-comm values
    configured (horizontally) in the community list is present in the prefix.
    There could be additional L-communities in the prefix that are not present
    in the L-comm list.
13. Verify that Match_ANY clause should pass only if at-least any one L-comm
    value configured(vertically) in large-community list, is present in prefixes.
14. Verify large-community lists operation in a route-map with match RegEx
    statements.
"""

import os
import sys
import pytest
import time

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# pylint: disable=C0413
# Import topogen and topotest helpers
# Import topoJson from lib, to create topology and initial configuration
from lib.topogen import Topogen, get_topogen

from lib.common_config import (
    start_topology,
    write_test_header,
    write_test_footer,
    reset_config_on_routers,
    create_route_maps,
    create_bgp_community_lists,
    verify_bgp_community,
    step,
    verify_create_community_list,
    delete_route_maps,
    verify_route_maps,
    create_static_routes,
    check_address_types,
    required_linux_kernel_version,
)
from lib.topolog import logger
from lib.bgp import verify_bgp_convergence, create_router_bgp, clear_bgp_and_verify
from lib.topojson import build_config_from_json


pytestmark = [pytest.mark.bgpd]


# Global variables
bgp_convergence = False

NETWORKS = {"ipv4": ["200.50.2.0/32"], "ipv6": ["1::1/128"]}


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    # Required linux kernel version for this suite to run.
    result = required_linux_kernel_version("4.15")
    if result is not True:
        pytest.skip("Kernel requirements are not met")

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/bgp_large_community_topo_2.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Checking BGP convergence
    global bgp_convergence, ADDR_TYPES

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Api call verify whether BGP is converged
    # Ipv4
    bgp_convergence = verify_bgp_convergence(tgen, topo)
    assert bgp_convergence is True, "setup_module :Failed \n Error:" " {}".format(
        bgp_convergence
    )
    ADDR_TYPES = check_address_types()

    logger.info("Running setup_module() done")


def teardown_module(mod):
    """
    Teardown the pytest environment

    * `mod`: module name
    """

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


#####################################################
#
#   Testcases
#
#####################################################


def test_create_bgp_standard_large_community_list(request):
    """
    Create standard large-community-list and verify it can permit
    or deny large community attribute only in the correct canonical
    format.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    reset_config_on_routers(tgen)

    step("Create srtandard large community list")
    input_dict = {
        "r4": {
            "bgp_community_lists": [
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "LC_1_STD",
                    "value": "2:1:1 2:1:2 1:2:3",
                    "large": True,
                },
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "LC_2_STD",
                    "value": "3:1:1 3:1:2",
                    "large": True,
                },
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify BGP large community is created")
    result = verify_create_community_list(tgen, input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create srtandard large community list with in-correct values")
    input_dict = {
        "r4": {
            "bgp_community_lists": [
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "LC_1_STD_ERR",
                    "value": "0:0:0",
                    "large": True,
                }
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    ## TODO should fail
    step("Verify BGP large community is created")
    result = verify_create_community_list(tgen, input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    write_test_footer(tc_name)


def test_create_bgp_expanded_large_community_list(request):
    """
    Create expanded large-community-list and verify it can permit
    or deny large community attribute both in the correct canonical
    format as well as REG_EX
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create expanded large community list")
    input_dict = {
        "r4": {
            "bgp_community_lists": [
                {
                    "community_type": "expanded",
                    "action": "permit",
                    "name": "LC_1_EXP",
                    "value": "1:1:200 1:2:* 3:2:1",
                    "large": True,
                }
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify BGP large community is created")
    result = verify_create_community_list(tgen, input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    write_test_footer(tc_name)


def test_modify_large_community_lists_referenced_by_rmap(request):
    """
    This test is to verify that we can modify a large-community-list
    is in use, add/remove attribute value and it takes immediate effect.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create standard large community list")
    input_dict_1 = {
        "r4": {
            "bgp_community_lists": [
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "LC_DEL",
                    "value": "1:2:1 1:3:1 2:1:1 2:2:2 3:3:3",
                    "large": True,
                }
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create route map")
    input_dict_2 = {
        "r1": {
            "route_maps": {
                "RM_R2_OUT": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {
                                "num": "1:2:1 1:3:1 2:10:1 3:3:3 4:4:4 5:5:5",
                                "action": "additive",
                            }
                        },
                    }
                ]
            }
        },
        "r4": {
            "route_maps": {
                "RM_R4_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {"large_comm_list": {"id": "LC_DEL", "delete": True}},
                    }
                ]
            }
        },
    }
    result = create_route_maps(tgen, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map and advertise networks")
    input_dict_3 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}],
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r1": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R2_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            },
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "advertise_networks": [{"network": "1::1/128"}],
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r1": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R2_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            },
                        }
                    },
                }
            }
        },
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
    }
    result = create_router_bgp(tgen, topo, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify Community-list")
    dut = "r4"
    input_dict_4 = {"largeCommunity": "2:10:1 4:4:4 5:5:5"}

    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_large_community_lists_with_rmap_apply_and_remove(request):
    """
    This test is to verify that large community attribute gets advertised when
    route-map is applied to a neighbor and cleared when route-map is removed
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create route map")
    input_dict_1 = {
        "r4": {
            "route_maps": {
                "RM_LC1": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {
                                "num": "200:200:1 200:200:10 200:200:20000",
                                "action": "additive",
                            }
                        },
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map and advertise networks")
    input_dict_2 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}]
                        }
                    },
                    "ipv6": {
                        "unicast": {"advertise_networks": [{"network": "1::1/128"}]}
                    },
                }
            }
        },
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r6": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_LC1", "direction": "out"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r6": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_LC1", "direction": "out"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
    }
    result = create_router_bgp(tgen, topo, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    dut = "r6"
    input_dict_4 = {"largeCommunity": "200:200:1 200:200:10 200:200:20000"}

    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Delete route map reference by community-list")
    input_dict_3 = {"r4": {"route_maps": ["RM_LC1"]}}
    result = delete_route_maps(tgen, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify route map is deleted")
    result = verify_route_maps(tgen, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    for adt in ADDR_TYPES:
        result = verify_bgp_community(
            tgen, adt, dut, NETWORKS[adt], input_dict_4, expected=False
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "largeCommunity is still present after deleting route-map \n Error: {}".format(
                tc_name, result
            )
        )

    write_test_footer(tc_name)


def test_duplicate_large_community_list_attributes_not_transitive(request):
    """
    This test is to verify that duplicate BGP Large Community values
    are NOT be transmitted.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create route map")
    input_dict_1 = {
        "r4": {
            "route_maps": {
                "RM_R4_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {
                                "num": "0:0:1 0:0:10 0:0:100 2:0:1 2:0:2 2:0:3"
                                " 2:0:4 2:0:5",
                                "action": "additive",
                            }
                        },
                    }
                ],
                "RM_R4_OUT": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {
                                "num": "0:0:1 0:0:10 0:0:10000 2:0:1 2:0:2",
                                "action": "additive",
                            }
                        },
                    }
                ],
            }
        }
    }
    result = create_route_maps(tgen, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map and advertise networks")
    input_dict_2 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}]
                        }
                    },
                    "ipv6": {
                        "unicast": {"advertise_networks": [{"network": "1::1/128"}]}
                    },
                }
            }
        },
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                },
                                "r6": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                },
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                },
                                "r6": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                },
                            }
                        }
                    },
                }
            }
        },
    }
    result = create_router_bgp(tgen, topo, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    dut = "r6"
    input_dict_4 = {
        "largeCommunity": "0:0:1 0:0:10 0:0:100 0:0:10000 2:0:1 2:0:2 2:0:3 2:0:4 2:0:5"
    }
    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_large_community_lists_with_rmap_set_none(request):
    """
    This test is to verify if we want to remove all the large-community
    attributes from a set of prefix we can set the value as NONE.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create route map")
    input_dict_1 = {
        "r4": {
            "route_maps": {
                "RM_R4_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {
                                "num": "0:0:1 0:0:10 0:0:100 2:0:1 2:0:2 2:0:3"
                                " 2:0:4",
                                "action": "additive",
                            }
                        },
                    }
                ]
            }
        },
        "r6": {
            "route_maps": {
                "RM_R6_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {"large_community": {"num": "none"}},
                    }
                ]
            }
        },
    }
    result = create_route_maps(tgen, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map")
    input_dict_2 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}]
                        }
                    },
                    "ipv6": {
                        "unicast": {"advertise_networks": [{"network": "1::1/128"}]}
                    },
                }
            }
        },
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
        "r6": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r6": {
                                            "route_maps": [
                                                {"name": "RM_R6_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r6": {
                                            "route_maps": [
                                                {"name": "RM_R6_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
    }
    result = create_router_bgp(tgen, topo, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify Community-list")
    dut = "r6"
    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], expected=False)
        assert result is not True, (
            "Testcase {} : Failed \n "
            "Community-list is still present \n Error: {}".format(tc_name, result)
        )

    write_test_footer(tc_name)


def test_lcomm_lists_with_redistribute_static_connected_rmap(request):
    """
    This test is to verify redistribute connected and static ipv4 routes
    in BGP process with a route-map appending/removing L-comm attributes.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("create static routes")
    input_dict = {
        "r1": {
            "static_routes": [
                {"network": "200.50.2.0/32", "next_hop": "10.0.0.6"},
                {"network": "1::1/128", "next_hop": "fd00:0:0:1::2"},
            ]
        }
    }
    result = create_static_routes(tgen, input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("redistribute static routes")
    input_dict_1 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "redistribute": [
                                {
                                    "redist_type": "static",
                                    "attribute": "route-map RM_R2_OUT",
                                },
                                {
                                    "redist_type": "connected",
                                    "attribute": "route-map RM_R2_OUT",
                                },
                            ]
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "redistribute": [
                                {
                                    "redist_type": "static",
                                    "attribute": "route-map RM_R2_OUT",
                                },
                                {
                                    "redist_type": "connected",
                                    "attribute": "route-map RM_R2_OUT",
                                },
                            ]
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create route map")
    input_dict_3 = {
        "r1": {
            "route_maps": {
                "RM_R2_OUT": [
                    {
                        "action": "permit",
                        "set": {"large_community": {"num": "55:55:55 555:555:555"}},
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list for static and connected ipv4 route on" " r2")

    input_dict_5 = {"largeCommunity": "55:55:55 555:555:555"}

    if "ipv4" in ADDR_TYPES:
        dut = "r2"
        networks = ["200.50.2.0/32", "1.0.1.17/32"]
        result = verify_bgp_community(tgen, "ipv4", dut, networks, input_dict_5)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        step("Verify large-community-list for static and connected ipv4 route" " on r4")
        dut = "r4"
        networks = ["200.50.2.0/32", "1.0.1.17/32"]
        result = verify_bgp_community(tgen, "ipv4", dut, networks, input_dict_5)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    if "ipv6" in ADDR_TYPES:
        step("Verify large-community-list for static and connected ipv6 route" " on r2")
        dut = "r2"
        networks = ["1::1/128", "2001:db8:f::1:17/128"]
        result = verify_bgp_community(tgen, "ipv6", dut, networks, input_dict_5)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        step("Verify large-community-list for static and connected ipv6 route" " on r4")
        dut = "r4"
        networks = ["1::1/128", "2001:db8:f::1:17/128"]
        result = verify_bgp_community(tgen, "ipv6", dut, networks, input_dict_5)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_large_community_lists_with_rmap_set_delete(request):
    """
    This test is to verify if we want to remove specific large-community
    values from a set of prefix we can make use of DELETE operation based
    on L-comm list
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("configure route_map")
    input_dict_2 = {
        "r6": {
            "bgp_community_lists": [
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "Test",
                    "value": "1:2:1 1:1:10 1:3:100",
                    "large": True,
                }
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create route map")
    input_dict_3 = {
        "r6": {
            "route_maps": {
                "RM_R6_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {"large_comm_list": {"id": "Test", "delete": True}},
                    }
                ]
            }
        },
        "r4": {
            "route_maps": {
                "RM_R4_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {
                                "num": "1:2:1 1:1:10 1:3:100 2:1:1 2:2:2 2:3:3"
                                " 2:4:4 2:5:5",
                                "action": "additive",
                            }
                        },
                    }
                ]
            }
        },
    }
    result = create_route_maps(tgen, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map and advertise networks")
    input_dict_4 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}]
                        }
                    },
                    "ipv6": {
                        "unicast": {"advertise_networks": [{"network": "1::1/128"}]}
                    },
                }
            }
        },
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
        "r6": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r6": {
                                            "route_maps": [
                                                {"name": "RM_R6_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r6": {
                                            "route_maps": [
                                                {"name": "RM_R6_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
    }
    result = create_router_bgp(tgen, topo, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    dut = "r6"
    input_dict_5 = {"largeCommunity": "2:1:1 2:2:2 2:3:3 2:4:4 2:5:5"}
    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], input_dict_5)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_large_community_lists_with_no_send_community(request):
    """
    This test is to verify if we want to remove specific large-community
    values from a set of prefix we can make use of DELETE operation based
    on L-comm list
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create route map")
    input_dict_2 = {
        "r5": {
            "route_maps": {
                "RM_R6_OUT": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {"num": "2:1:1 2:2:2 2:3:3 2:4:4 2:5:5"}
                        },
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map and advertise networks")
    input_dict_3 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}]
                        }
                    },
                    "ipv6": {
                        "unicast": {"advertise_networks": [{"network": "1::1/128"}]}
                    },
                }
            }
        },
        "r5": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r6": {
                                    "dest_link": {
                                        "r5": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R6_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r6": {
                                    "dest_link": {
                                        "r5": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R6_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
    }
    result = create_router_bgp(tgen, topo, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    dut = "r6"
    input_dict_4 = {"largeCommunity": "2:1:1 2:2:2 2:3:3 2:4:4 2:5:5"}
    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Configure neighbor for no-send-community")
    input_dict_5 = {
        "r5": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r6": {
                                    "dest_link": {"r5": {"no_send_community": "large"}}
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r6": {
                                    "dest_link": {"r5": {"no_send_community": "large"}}
                                }
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify Community-list")
    for adt in ADDR_TYPES:
        result = verify_bgp_community(
            tgen, adt, dut, NETWORKS[adt], input_dict_4, expected=False
        )
        assert result is not True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_create_large_community_lists_with_no_attribute_values(request):
    """
    This test is to verify that large-community lists can not be
    configured without providing specific L-community values
    (for match/delete operation in a route-map).
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create standard large commumity-list")
    input_dict_1 = {
        "r5": {
            "bgp_community_lists": [
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "Test1",
                    "large": True,
                }
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict_1)
    assert result is not True, "Testcase {} : Failed \n Error: {}".format(
        tc_name, result
    )

    write_test_footer(tc_name)


def test_large_community_lists_with_rmap_match_exact(request):
    """
    This test is to verify that Match_EXACT clause should pass
    only if all of the L-comm values configured (horizontally)
    in the community list is present in the prefix. There must
    be no additional L-communities in the prefix.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create route map")
    input_dict_2 = {
        "r2": {
            "route_maps": {
                "RM_R4_OUT": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {"num": "2:1:1 2:2:2 2:3:3 2:4:4 2:5:5"}
                        },
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map and advertise networks")
    input_dict_3 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}]
                        }
                    },
                    "ipv6": {
                        "unicast": {"advertise_networks": [{"network": "1::1/128"}]}
                    },
                }
            }
        },
        "r2": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r2": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r2": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
    }

    result = create_router_bgp(tgen, topo, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create standard large commumity-list")
    input_dict_4 = {
        "r4": {
            "bgp_community_lists": [
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "EXACT",
                    "value": "2:1:1 2:2:2 2:3:3 2:4:4 2:5:5",
                    "large": True,
                }
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify BGP large community is created")
    result = verify_create_community_list(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create route map")
    input_dict_5 = {
        "r4": {
            "route_maps": {
                "RM_R4_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "match": {
                            "large-community-list": ["EXACT"],
                            "match_exact": True,
                        },
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map")
    input_dict_6 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_6)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    dut = "r4"
    input_dict_4 = {"largeCommunity": "2:1:1 2:2:2 2:3:3 2:4:4 2:5:5"}
    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_large_community_lists_with_rmap_match_all(request):
    """
    This test is to verify that Match_ALL clause should pass
    only if ALL of the L-comm values configured (horizontally)
    in the community list are present in the prefix. There
    could be additional L-communities in the prefix that are
    not present in the L-comm list.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create route map")
    input_dict_2 = {
        "r2": {
            "route_maps": {
                "RM_R4_OUT": [
                    {
                        "action": "permit",
                        "set": {
                            "large_community": {
                                "num": "1:1:1 1:2:3 2:1:1 2:2:2 2:3:3 2:4:4 2:5:5"
                            }
                        },
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map")
    input_dict_3 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}]
                        }
                    },
                    "ipv6": {
                        "unicast": {"advertise_networks": [{"network": "1::1/128"}]}
                    },
                }
            }
        },
        "r2": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r2": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r2": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
    }
    result = create_router_bgp(tgen, topo, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create standard large commumity-list")
    input_dict_4 = {
        "r3": {
            "bgp_community_lists": [
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "ALL",
                    "value": "2:1:1 2:2:2 2:3:3 2:4:4 2:5:5",
                    "large": True,
                }
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify BGP large community is created")
    result = verify_create_community_list(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create route map")
    input_dict_5 = {
        "r4": {
            "route_maps": {
                "RM_R4_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "match": {"large-community-list": {"id": "ALL"}},
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map")
    input_dict_6 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_6)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    dut = "r4"
    input_dict_4 = {"largeCommunity": "1:1:1 1:2:3 2:1:1 2:2:2 2:3:3 2:4:4 2:5:5"}
    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], input_dict_4)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_large_community_lists_with_rmap_match_any(request):
    """
    This test is to verify that Match_ANY clause should pass
    only if at-least any one L-comm value configured(vertically)
    in large-community list, is present in prefixes.
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create route map")
    input_dict_2 = {
        "r2": {
            "route_maps": {
                "RM_R4_OUT": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {"num": "2:1:1 2:2:2 2:3:3 2:4:4 2:5:5"}
                        },
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map")
    input_dict_3 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}]
                        }
                    },
                    "ipv6": {
                        "unicast": {"advertise_networks": [{"network": "1::1/128"}]}
                    },
                }
            }
        },
        "r2": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r2": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r2": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
    }
    result = create_router_bgp(tgen, topo, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create standard large commumity-list")
    input_dict_4 = {
        "r4": {
            "bgp_community_lists": [
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "ANY",
                    "value": "2:1:1",
                    "large": True,
                },
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "ANY",
                    "value": "2:2:1",
                    "large": True,
                },
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "ANY",
                    "value": "2:3:1",
                    "large": True,
                },
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "ANY",
                    "value": "2:4:1",
                    "large": True,
                },
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify BGP large community is created")
    result = verify_create_community_list(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create route map")
    input_dict_5 = {
        "r4": {
            "route_maps": {
                "RM_R4_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "match": {"large-community-list": {"id": "ANY"}},
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map")
    input_dict_6 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_6)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    dut = "r4"
    input_dict_7 = {"largeCommunity": "2:1:1 2:2:2 2:3:3 2:4:4 2:5:5"}
    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], input_dict_7)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


def test_large_community_lists_with_rmap_match_regex(request):
    """
    This test is to verify large-community lists" operation in a route-map
    with match RegEx statements. Match clause should pass only if the
    complete string of L-comm values are matched
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    reset_config_on_routers(tgen)

    step("Create route map")
    input_dict_2 = {
        "r2": {
            "route_maps": {
                "RM_R4_OUT": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "set": {
                            "large_community": {
                                "num": "1:1:1 1:1:2 2:1:3 2:1:4 2:1:5",
                            },
                            "community": {"num": "1:1 1:2 1:3 1:4 1:5"},
                        },
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_2)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map")
    input_dict_3 = {
        "r1": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "advertise_networks": [{"network": "200.50.2.0/32"}]
                        }
                    },
                    "ipv6": {
                        "unicast": {"advertise_networks": [{"network": "1::1/128"}]}
                    },
                }
            }
        },
        "r2": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r2": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r4": {
                                    "dest_link": {
                                        "r2": {
                                            "route_maps": [
                                                {
                                                    "name": "RM_R4_OUT",
                                                    "direction": "out",
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        },
    }
    result = create_router_bgp(tgen, topo, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create standard large commumity-list")
    input_dict_4 = {
        "r4": {
            "bgp_community_lists": [
                {
                    "community_type": "standard",
                    "action": "permit",
                    "name": "ALL",
                    "value": "1:1:1 2:1:3 2:1:4 2:1:5",
                    "large": True,
                },
                {
                    "community_type": "expanded",
                    "action": "permit",
                    "name": "EXP_ALL",
                    "value": "1:1:1 2:1:[3-5]",
                    "large": True,
                },
            ]
        }
    }
    result = create_bgp_community_lists(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify BGP large community is created")
    result = verify_create_community_list(tgen, input_dict_4)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create route map")
    input_dict_5 = {
        "r4": {
            "route_maps": {
                "RM_R4_IN": [
                    {
                        "action": "permit",
                        "seq_id": "10",
                        "match": {
                            "large_community_list": {
                                "id": "ALL",
                            },
                        },
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Configure neighbor for route map")
    input_dict_6 = {
        "r4": {
            "bgp": {
                "address_family": {
                    "ipv4": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "ipv6": {
                        "unicast": {
                            "neighbor": {
                                "r2": {
                                    "dest_link": {
                                        "r4": {
                                            "route_maps": [
                                                {"name": "RM_R4_IN", "direction": "in"}
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    },
                }
            }
        }
    }
    result = create_router_bgp(tgen, topo, input_dict_6)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    dut = "r4"
    input_dict_7 = {"largeCommunity": "1:1:1 1:1:2 2:1:3 2:1:4 2:1:5"}
    for adt in ADDR_TYPES:
        result = verify_bgp_community(tgen, adt, dut, NETWORKS[adt], input_dict_7)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    step("Delete route map reference by community-list")
    input_dict_3 = {"r4": {"route_maps": ["RM_R4_IN"]}}
    result = delete_route_maps(tgen, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    result = verify_route_maps(tgen, input_dict_3)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Create route map")
    input_dict_5 = {
        "r4": {
            "route_maps": {
                "RM_R4_IN": [
                    {
                        "action": "permit",
                        "seq_id": "20",
                        "match": {
                            "large_community_list": {
                                "id": "EXP_ALL",
                            },
                        },
                    }
                ]
            }
        }
    }
    result = create_route_maps(tgen, input_dict_5)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("clear ip bgp")
    result = clear_bgp_and_verify(tgen, topo, "r4")
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("Verify large-community-list")
    dut = "r4"
    input_dict_7 = {"largeCommunity": "1:1:1 1:1:2 2:1:3 2:1:4 2:1:5"}
    for adt in ADDR_TYPES:
        result = verify_bgp_community(
            tgen, adt, dut, NETWORKS[adt], input_dict_7, expected=False
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "largeCommunity is still present \n Error: {}".format(tc_name, result)
        )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2021 by
# Donatas Abraitis <donatas.abraitis@gmail.com>
#

"""
Test if we can match BGP prefixes by next-hop which is
specified by an IPv6 Access-list, prefix-list or just an address.
"""

import os
import sys
import json
import pytest
import functools

pytestmark = [pytest.mark.bgpd]

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen


def build_topo(tgen):
    for routern in range(1, 3):
        tgen.add_router("r{}".format(routern))

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_route_map_match_ipv6_next_hop_access_list():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    router = tgen.gears["r1"]

    def _bgp_converge(router):
        output = json.loads(router.vtysh_cmd("show ipv6 route json"))
        expected = {
            "2001:db8:1::1/128": [
                {
                    "communities": "65002:1",
                }
            ],
            "2001:db8:2::1/128": [
                {
                    "communities": "65002:2",
                }
            ],
            "2001:db8:3::1/128": [
                {
                    "communities": "65002:3",
                }
            ],
            "2001:db8:4::1/128": [
                {
                    "communities": "65002:4",
                }
            ],
            "2001:db8:5::1/128": [
                {
                    "communities": "65002:5",
                }
            ],
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_converge, router)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Can't match routes using ipv6 next-hop access-list"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_srv6_route.py
#
# Copyright 2023 6WIND S.A.
# Dmytro Shytyi <dmytro.shytyi@6wind.com>
#

"""
test_srv6_route.py:
Test for SRv6 static route on zebra
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

pytestmark = [pytest.mark.staticd]


def open_json_file(filename):
    try:
        with open(filename, "r") as f:
            return json.load(f)
    except IOError:
        assert False, "Could not read file {}".format(filename)


def setup_module(mod):
    tgen = Topogen({None: "r1"}, mod.__name__)
    tgen.start_topology()
    for rname, router in tgen.routers().items():
        router.run("/bin/bash {}/{}/setup.sh".format(CWD, rname))
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_MGMTD, os.path.join(CWD, "{}/mgmtd.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_STATIC, os.path.join(CWD, "{}/staticd.conf".format(rname))
        )
    tgen.start_router()


def teardown_module():
    tgen = get_topogen()
    tgen.stop_topology()


def test_srv6_static_route():
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)
    router = tgen.gears["r1"]

    def _check_srv6_static_route(router, expected_route_file):
        logger.info("checking zebra srv6 static route with multiple segs status")
        output = json.loads(router.vtysh_cmd("show ipv6 route static json"))
        expected = open_json_file("{}/{}".format(CWD, expected_route_file))
        return topotest.json_cmp(output, expected)

    def check_srv6_static_route(router, expected_file):
        func = functools.partial(_check_srv6_static_route, router, expected_file)
        _, result = topotest.run_and_expect(func, None, count=15, wait=1)
        assert result is None, "Failed"

    # FOR DEVELOPER:
    # If you want to stop some specific line and start interactive shell,
    # please use tgen.mininet_cli() to start it.

    logger.info("Test for srv6 route configuration")
    check_srv6_static_route(router, "expected_srv6_route.json")


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: ISC
#
# Copyright (c) 2021, LabN Consulting, L.L.C.
# Copyright (c) 2019-2020 by
# Donatas Abraitis <donatas.abraitis@gmail.com>
#
# noqa: E501
#
"""
Test static route functionality
"""
import pytest
from lib.topogen import Topogen
from oper import check_kernel_32, do_oper_test

pytestmark = [pytest.mark.staticd, pytest.mark.mgmtd]


@pytest.fixture(scope="module")
def tgen(request):
    "Setup/Teardown the environment and provide tgen argument to tests"

    topodef = {"s1": ("r1",), "s2": ("r1",)}

    tgen = Topogen(topodef, request.module.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for rname, router in router_list.items():
        # Setup VRF red
        router.net.add_l3vrf("red", 10)
        router.net.add_loop("lo-red")
        router.net.attach_iface_to_l3vrf("lo-red", "red")
        router.net.attach_iface_to_l3vrf(rname + "-eth1", "red")
        router.load_frr_config("frr-simple.conf")

    tgen.start_router()
    yield tgen
    tgen.stop_topology()


def test_oper_simple(tgen):
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    query_results = [
        (
            # Non-key query with key specific selection
            '/frr-interface:lib/interface[name="r1-eth0"]/vrf',
            "simple-results/result-intf-eth0-vrf.json",
        ),
        # Test machines will have different sets of interfaces so the test results will
        # vary and need to be generated dynamically before this test is re-enabled
        # (
        #     # Key query on generic list
        #     "/frr-interface:lib/interface/name",
        #     "simple-results/result-intf-name.json",
        # ),
        (
            # Key query with key specific selection
            '/frr-interface:lib/interface[name="r1-eth0"]/name',
            "simple-results/result-intf-eth0-name.json",
        ),
        ("/frr-vrf:lib", "simple-results/result-lib.json"),
        ("/frr-vrf:lib/vrf", "simple-results/result-lib-vrf-nokey.json"),
        (
            '/frr-vrf:lib/vrf[name="default"]',
            "simple-results/result-lib-vrf-default.json",
        ),
        ('/frr-vrf:lib/vrf[name="red"]', "simple-results/result-lib-vrf-red.json"),
        (
            '/frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra',
            "simple-results/result-lib-vrf-zebra.json",
        ),
        (
            '/frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs',
            "simple-results/result-lib-vrf-zebra-ribs.json",
        ),
        (
            '/frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/rib',
            "simple-results/result-ribs-rib-nokeys.json",
        ),
        (
            '/frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/'
            'rib[afi-safi-name="frr-routing:ipv4-unicast"][table-id="254"]',
            "simple-results/result-ribs-rib-ipv4-unicast.json",
        ),
        (
            '/frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/'
            'rib[afi-safi-name="frr-routing:ipv4-unicast"][table-id="254"]/route',
            "simple-results/result-ribs-rib-route-nokey.json",
        ),
        (
            '/frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/'
            'rib[afi-safi-name="frr-routing:ipv4-unicast"][table-id="254"]/'
            'route[prefix="1.1.1.0/24"]',
            "simple-results/result-ribs-rib-route-prefix.json",
        ),
        # Missing entry
        (
            '/frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/'
            'rib[afi-safi-name="frr-routing:ipv4-unicast"][table-id="254"]/'
            'route[prefix="1.1.0.0/24"]',
            "simple-results/result-empty.json",
        ),
        # Leaf reference
        (
            '/frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/'
            'rib[afi-safi-name="frr-routing:ipv4-unicast"][table-id="254"]/'
            'route[prefix="1.1.1.0/24"]/route-entry[protocol="connected"]/metric',
            "simple-results/result-singleton-metric.json",
        ),
        (
            '/frr-interface:lib/interface[name="r1-eth0"]',
            "simple-results/result-intf-eth0-with-config.json",
            "with-config",
        ),
        (
            '/frr-interface:lib/interface[name="r1-eth0"]',
            "simple-results/result-intf-eth0-only-config.json",
            "only-config",
        ),
        (
            "/frr-interface:lib/interface/description",
            "simple-results/result-intf-description.json",
            "with-config",
        ),
        (
            '/frr-interface:lib/interface[name="r1-eth0"]',
            "simple-results/result-intf-eth0-exact.json",
            "exact",
        ),
        (
            '/frr-interface:lib/interface[name="r1-eth0"]/description',
            "simple-results/result-intf-eth0-description-exact.json",
            "with-config exact",
        ),
        # Interface state
        (
            '/frr-interface:lib/interface[name="r1-eth0"]/state',
            "simple-results/result-intf-state.json",
        ),
        (
            '/frr-interface:lib/interface[name="r1-eth0"]/state/mtu',
            "simple-results/result-intf-state-mtu.json",
        ),
        # with-defaults
        (
            '/frr-interface:lib/interface[name="r1-eth0"]/frr-zebra:zebra/evpn-mh',
            "simple-results/result-intf-eth0-wd-explicit.json",
            "with-config exact",
        ),
        (
            '/frr-interface:lib/interface[name="r1-eth0"]/frr-zebra:zebra/evpn-mh',
            (
                # Output is different between libyang2 and libyang3+
                "simple-results/result-intf-eth0-wd-trim.json",
                "simple-results/result-intf-eth0-wd-trim-empty-label.json",
            ),
            "with-config exact with-defaults trim",
        ),
        (
            '/frr-interface:lib/interface[name="r1-eth0"]/frr-zebra:zebra/evpn-mh',
            "simple-results/result-intf-eth0-wd-all.json",
            "with-config exact with-defaults all",
        ),
        (
            '/frr-interface:lib/interface[name="r1-eth0"]/frr-zebra:zebra/evpn-mh',
            "simple-results/result-intf-eth0-wd-all-tag.json",
            "with-config exact with-defaults all-tag",
        ),
    ]

    r1 = tgen.gears["r1"].net
    check_kernel_32(r1, "11.11.11.11", 1, "")
    do_oper_test(tgen, query_results)


to_gen_new_results = """
scriptdir=~chopps/w/frr/tests/topotests/mgmt_oper
resdir=${scriptdir}/simple-results
vtysh -c 'show mgmt get-data /frr-vrf:lib'      > ${resdir}/result-lib.json
vtysh -c 'show mgmt get-data /frr-vrf:lib/vrf'  > ${resdir}/result-lib-vrf-nokey.json
vtysh -c 'show mgmt get-data /frr-vrf:lib/vrf[name="default"]'  > ${resdir}/result-lib-vrf-default.json
vtysh -c 'show mgmt get-data /frr-vrf:lib/vrf[name="red"]'      > ${resdir}/result-lib-vrf-red.json
vtysh -c 'show mgmt get-data /frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra'          > ${resdir}/result-lib-vrf-zebra.json
vtysh -c 'show mgmt get-data /frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs'     > ${resdir}/result-lib-vrf-zebra-ribs.json
vtysh -c 'show mgmt get-data /frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/rib' > ${resdir}/result-ribs-rib-nokeys.json
vtysh -c 'show mgmt get-data /frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/rib[afi-safi-name="frr-routing:ipv4-unicast"][table-id="254"]' > ${resdir}/result-ribs-rib-ipv4-unicast.json
vtysh -c 'show mgmt get-data /frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/rib[afi-safi-name="frr-routing:ipv4-unicast"][table-id="254"]/route' > ${resdir}/result-ribs-rib-route-nokey.json

vtysh -c 'show mgmt get-data /frr-interface:lib/interface[name="r1-eth0"]/state' > ${resdir}/result-intf-state.json
vtysh -c 'show mgmt get-data /frr-interface:lib/interface[name="r1-eth0"]/state/mtu' > ${resdir}/result-intf-state-mtu.json

for f in ${resdir}/result-*; do
   sed -i -e 's/"uptime": ".*"/"uptime": "rubout"/;s/"id": [0-9][0-9]*/"id": "rubout"/' $f
   sed -i -e 's/"phy-address": ".*"/"phy-address": "rubout"/' $f
   sed -i -e 's/"if-index": [0-9][0-9]*/"if-index": "rubout"/' $f
   sed -i -e 's,"vrf": "[0-9]*","vrf": "rubout",' $f
done
"""  # noqa: 501

# Example commands:
# show mgmt get-data /frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/rib[afi-safi-name="frr-routing:ipv4-unicast"][table-id="254"]/route[prefix="1.1.0.0/24"] # noqa: E501
# show mgmt get-data /frr-vrf:lib/vrf[name="default"]/frr-zebra:zebra/ribs/rib[afi-safi-name="frr-routing:ipv4-unicast"][table-id="254"]/route[prefix="1.1.1.0/24"] # noqa: E501



# SPDX-License-Identifier: ISC
#
# Modified work Copyright (c) 2019 by VMware, Inc. ("VMware")
# Original work Copyright (c) 2018 by Network Device Education
# Foundation, Inc. ("NetDEF")
#

import json
import ipaddress
import os
from collections import OrderedDict
from copy import deepcopy
from re import search as re_search

import pytest

from lib.bgp import create_router_bgp
from lib.common_config import (
    create_bgp_community_lists,
    create_interfaces_cfg,
    create_prefix_lists,
    create_route_maps,
    create_static_routes,
    create_vrf_cfg,
    load_config_to_routers,
    start_topology,
    topo_daemons,
    number_to_column,
)
from lib.ospf import create_router_ospf
from lib.pim import (
    create_igmp_config,
    create_pim_config,
    create_mld_config,
)
from lib.topolog import logger


def build_topo_from_json(tgen, topo=None):
    """
    Reads configuration from JSON file. Adds routers, creates interface
    names dynamically and link routers as defined in JSON to create
    topology. Assigns IPs dynamically to all interfaces of each router.
    * `tgen`: Topogen object
    * `topo`: json file data, or use tgen.json_topo if None
    """
    if topo is None:
        topo = tgen.json_topo

    router_list = sorted(
        topo["routers"].keys(), key=lambda x: int(re_search(r"\d+", x).group(0))
    )

    switch_list = []
    if "switches" in topo:
        switch_list = sorted(
            topo["switches"].keys(), key=lambda x: int(re_search(r"\d+", x).group(0))
        )

    listRouters = sorted(router_list[:])
    listSwitches = sorted(switch_list[:])
    listAllRouters = deepcopy(listRouters)
    dictSwitches = {}

    for routerN in router_list:
        logger.info("Topo: Add router {}".format(routerN))
        tgen.add_router(routerN)

    for switchN in switch_list:
        logger.info("Topo: Add switch {}".format(switchN))
        dictSwitches[switchN] = tgen.add_switch(switchN)

    if "ipv4base" in topo:
        ipv4Next = ipaddress.IPv4Address(topo["link_ip_start"]["ipv4"])
        ipv4Step = 2 ** (32 - topo["link_ip_start"]["v4mask"])
        if topo["link_ip_start"]["v4mask"] < 32:
            ipv4Next += 1
    if "ipv6base" in topo:
        ipv6Next = ipaddress.IPv6Address(topo["link_ip_start"]["ipv6"])
        ipv6Step = 2 ** (128 - topo["link_ip_start"]["v6mask"])
        if topo["link_ip_start"]["v6mask"] < 127:
            ipv6Next += 1
    for router in listRouters:
        topo["routers"][router]["nextIfname"] = 0

    router_count = 0
    while listRouters != []:
        curRouter = listRouters.pop(0)
        # Physical Interfaces
        if "links" in topo["routers"][curRouter]:
            for destRouterLink, data in sorted(
                topo["routers"][curRouter]["links"].items()
            ):
                currRouter_lo_json = topo["routers"][curRouter]["links"][destRouterLink]
                # Loopback interfaces
                if "type" in data and data["type"] == "loopback":
                    router_count += 1
                    if (
                        "ipv4" in currRouter_lo_json
                        and currRouter_lo_json["ipv4"] == "auto"
                    ):
                        currRouter_lo_json["ipv4"] = "{}{}.{}/{}".format(
                            topo["lo_prefix"]["ipv4"],
                            router_count,
                            number_to_column(curRouter),
                            topo["lo_prefix"]["v4mask"],
                        )
                    if (
                        "ipv6" in currRouter_lo_json
                        and currRouter_lo_json["ipv6"] == "auto"
                    ):
                        currRouter_lo_json["ipv6"] = "{}{}:{}/{}".format(
                            topo["lo_prefix"]["ipv6"],
                            router_count,
                            number_to_column(curRouter),
                            topo["lo_prefix"]["v6mask"],
                        )

                if "-" in destRouterLink:
                    # Spliting and storing destRouterLink data in tempList
                    tempList = destRouterLink.split("-")

                    # destRouter
                    destRouter = tempList.pop(0)

                    # Current Router Link
                    tempList.insert(0, curRouter)
                    curRouterLink = "-".join(tempList)
                else:
                    destRouter = destRouterLink
                    curRouterLink = curRouter

                if destRouter in listRouters:
                    currRouter_link_json = topo["routers"][curRouter]["links"][
                        destRouterLink
                    ]
                    destRouter_link_json = topo["routers"][destRouter]["links"][
                        curRouterLink
                    ]

                    # Assigning name to interfaces
                    currRouter_link_json["interface"] = "{}-{}-eth{}".format(
                        curRouter, destRouter, topo["routers"][curRouter]["nextIfname"]
                    )
                    destRouter_link_json["interface"] = "{}-{}-eth{}".format(
                        destRouter, curRouter, topo["routers"][destRouter]["nextIfname"]
                    )

                    # add link interface
                    destRouter_link_json["peer-interface"] = "{}-{}-eth{}".format(
                        curRouter, destRouter, topo["routers"][curRouter]["nextIfname"]
                    )
                    currRouter_link_json["peer-interface"] = "{}-{}-eth{}".format(
                        destRouter, curRouter, topo["routers"][destRouter]["nextIfname"]
                    )

                    topo["routers"][curRouter]["nextIfname"] += 1
                    topo["routers"][destRouter]["nextIfname"] += 1

                    # Linking routers to each other as defined in JSON file
                    tgen.gears[curRouter].add_link(
                        tgen.gears[destRouter],
                        topo["routers"][curRouter]["links"][destRouterLink][
                            "interface"
                        ],
                        topo["routers"][destRouter]["links"][curRouterLink][
                            "interface"
                        ],
                    )

                    # IPv4
                    if "ipv4" in currRouter_link_json:
                        if currRouter_link_json["ipv4"] == "auto":
                            currRouter_link_json["ipv4"] = "{}/{}".format(
                                ipv4Next, topo["link_ip_start"]["v4mask"]
                            )
                            destRouter_link_json["ipv4"] = "{}/{}".format(
                                ipv4Next + 1, topo["link_ip_start"]["v4mask"]
                            )
                            ipv4Next += ipv4Step
                    # IPv6
                    if "ipv6" in currRouter_link_json:
                        if currRouter_link_json["ipv6"] == "auto":
                            currRouter_link_json["ipv6"] = "{}/{}".format(
                                ipv6Next, topo["link_ip_start"]["v6mask"]
                            )
                            destRouter_link_json["ipv6"] = "{}/{}".format(
                                ipv6Next + 1, topo["link_ip_start"]["v6mask"]
                            )
                            ipv6Next = ipaddress.IPv6Address(int(ipv6Next) + ipv6Step)

            logger.debug(
                "Generated link data for router: %s\n%s",
                curRouter,
                json.dumps(
                    topo["routers"][curRouter]["links"], indent=4, sort_keys=True
                ),
            )

    switch_count = 0
    add_switch_to_topo = []
    while listSwitches != []:
        curSwitch = listSwitches.pop(0)
        # Physical Interfaces
        if "links" in topo["switches"][curSwitch]:
            for destRouterLink, data in sorted(
                topo["switches"][curSwitch]["links"].items()
            ):
                # Loopback interfaces
                if "dst_node" in data:
                    destRouter = data["dst_node"]

                elif "-" in destRouterLink:
                    # Spliting and storing destRouterLink data in tempList
                    tempList = destRouterLink.split("-")
                    # destRouter
                    destRouter = tempList.pop(0)
                else:
                    destRouter = destRouterLink

                if destRouter in listAllRouters:
                    topo["routers"][destRouter]["links"][curSwitch] = deepcopy(
                        topo["switches"][curSwitch]["links"][destRouterLink]
                    )

                    # Assigning name to interfaces
                    topo["routers"][destRouter]["links"][curSwitch][
                        "interface"
                    ] = "{}-{}-eth{}".format(
                        destRouter,
                        curSwitch,
                        topo["routers"][destRouter]["nextIfname"],
                    )

                    topo["switches"][curSwitch]["links"][destRouter][
                        "interface"
                    ] = "{}-{}-eth{}".format(
                        curSwitch,
                        destRouter,
                        topo["routers"][destRouter]["nextIfname"],
                    )

                    topo["routers"][destRouter]["nextIfname"] += 1

                    # Add links
                    dictSwitches[curSwitch].add_link(
                        tgen.gears[destRouter],
                        topo["switches"][curSwitch]["links"][destRouter]["interface"],
                        topo["routers"][destRouter]["links"][curSwitch]["interface"],
                    )

                    # IPv4
                    if "ipv4" in topo["routers"][destRouter]["links"][curSwitch]:
                        if (
                            topo["routers"][destRouter]["links"][curSwitch]["ipv4"]
                            == "auto"
                        ):
                            topo["routers"][destRouter]["links"][curSwitch][
                                "ipv4"
                            ] = "{}/{}".format(
                                ipv4Next, topo["link_ip_start"]["v4mask"]
                            )
                            ipv4Next += 1
                    # IPv6
                    if "ipv6" in topo["routers"][destRouter]["links"][curSwitch]:
                        if (
                            topo["routers"][destRouter]["links"][curSwitch]["ipv6"]
                            == "auto"
                        ):
                            topo["routers"][destRouter]["links"][curSwitch][
                                "ipv6"
                            ] = "{}/{}".format(
                                ipv6Next, topo["link_ip_start"]["v6mask"]
                            )
                            ipv6Next = ipaddress.IPv6Address(int(ipv6Next) + ipv6Step)

            logger.debug(
                "Generated link data for router: %s\n%s",
                curRouter,
                json.dumps(
                    topo["routers"][curRouter]["links"], indent=4, sort_keys=True
                ),
            )


def linux_intf_config_from_json(tgen, topo=None):
    """Configure interfaces from linux based on topo."""
    if topo is None:
        topo = tgen.json_topo

    routers = topo["routers"]
    for rname in routers:
        router = tgen.net[rname]
        links = routers[rname]["links"]
        for rrname in links:
            link = links[rrname]
            if rrname == "lo":
                lname = "lo"
            else:
                lname = link["interface"]
            if "ipv4" in link:
                router.cmd_raises("ip addr add {} dev {}".format(link["ipv4"], lname))
            if "ipv6" in link:
                router.cmd_raises(
                    "ip -6 addr add {} dev {}".format(link["ipv6"], lname)
                )


def build_config_from_json(tgen, topo=None, save_bkup=True):
    """
    Reads initial configuraiton from JSON for each router, builds
    configuration and loads its to router.

    * `tgen`: Topogen object
    * `topo`: json file data, or use tgen.json_topo if None
    """

    func_dict = OrderedDict(
        [
            ("vrfs", create_vrf_cfg),
            ("ospf", create_router_ospf),
            ("links", create_interfaces_cfg),
            ("static_routes", create_static_routes),
            ("prefix_lists", create_prefix_lists),
            ("bgp_community_list", create_bgp_community_lists),
            ("route_maps", create_route_maps),
            ("pim", create_pim_config),
            ("igmp", create_igmp_config),
            ("mld", create_mld_config),
            ("bgp", create_router_bgp),
        ]
    )

    if topo is None:
        topo = tgen.json_topo

    data = topo["routers"]
    for func_type in func_dict.keys():
        logger.info("Checking for {} configuration in input data".format(func_type))

        func_dict.get(func_type)(tgen, data, build=True)

    routers = sorted(topo["routers"].keys())
    result = load_config_to_routers(tgen, routers, save_bkup)
    if not result:
        logger.info("build_config_from_json: failed to configure topology")
        assert False

    logger.info(
        "Built config now clearing ospf neighbors as that router-id might not be what is used"
    )
    for ospf in ["ospf", "ospf6"]:
        for router in data:
            if ospf not in data[router]:
                continue

            r = tgen.gears[router]
            if ospf == "ospf":
                r.vtysh_cmd("clear ip ospf process")
            else:
                r.vtysh_cmd("clear ipv6 ospf6 process")


def create_tgen_from_json(testfile, json_file=None):
    """Create a topogen object given a testfile.

    - `testfile`  : The path to the testfile.
    - `json_file` : The path to the json config file. If None the pathname is derived
      from the `testfile` first by trying to replace `.py` by `.json` and if that isn't
      present then by removing `test_` prefix as well.
    """
    from lib.topogen import Topogen  # Topogen imports this module too

    thisdir = os.path.dirname(os.path.realpath(testfile))
    basename = os.path.basename(testfile)
    logger.debug("starting standard JSON based module setup for %s", basename)

    assert basename.startswith("test_")
    assert basename.endswith(".py")
    json_file = os.path.join(thisdir, basename[:-3] + ".json")
    if not os.path.exists(json_file):
        json_file = os.path.join(thisdir, basename[5:-3] + ".json")
        assert os.path.exists(json_file)
    with open(json_file, "r") as topof:
        topo = json.load(topof)

    # Create topology
    tgen = Topogen(lambda tgen: build_topo_from_json(tgen, topo), basename[:-3])
    tgen.json_topo = topo
    return tgen


def setup_module_from_json(testfile, json_file=None):
    """Do the standard module setup for JSON based test.

    * `testfile` : The path to the testfile. The name is used to derive the json config
      file name as well (removing `test_` prefix and replacing `.py` suffix with `.json`
    """
    # Create topology object
    tgen = create_tgen_from_json(testfile, json_file)

    # Start routers (and their daemons)
    start_topology(tgen)

    # Configure routers
    build_config_from_json(tgen)
    assert not tgen.routers_have_failure()

    return tgen



#!/usr/bin/env python
# -*- coding: utf-8 eval: (blacken-mode 1) -*-
# SPDX-License-Identifier: ISC
#
# test_ospf_topo2.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2017 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_ospf_topo2.py: Test correct route removal.

Proofs the following issue:
https://github.com/FRRouting/frr/issues/14488

"""

import ipaddress
import json
import pytest
import sys
import time

from lib.topogen import Topogen


pytestmark = [
    pytest.mark.ospf6d,
    pytest.mark.ospfd,
]


def build_topo(tgen):
    """Build the topology used by all tests below."""

    # Create 4 routers
    r1 = tgen.add_router("r1")
    r2 = tgen.add_router("r2")
    r3 = tgen.add_router("r3")
    r4 = tgen.add_router("r4")

    # The r1/r2 and r3/r4 router pairs have two connections each
    tgen.add_link(r1, r2, ifname1="eth1", ifname2="eth1")
    tgen.add_link(r1, r2, ifname1="eth2", ifname2="eth2")
    tgen.add_link(r3, r4, ifname1="eth2", ifname2="eth2")
    tgen.add_link(r3, r4, ifname1="eth3", ifname2="eth3")

    # The r1/r4 and r2/r3 router pairs have one connection each
    tgen.add_link(r1, r4, ifname1="eth3", ifname2="eth1")
    tgen.add_link(r2, r3, ifname1="eth3", ifname2="eth1")


@pytest.fixture(scope="function")
def tgen(request):
    """Setup/Teardown the environment and provide tgen argument to tests.

    Do this once per function as some of the tests will leave the router
    in an unclean state.

    """

    tgen = Topogen(build_topo, request.module.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, router in router_list.items():
        router.load_frr_config("frr.conf")

    tgen.start_router()

    yield tgen

    tgen.stop_topology()


def ospf_neighbors(router, ip_version):
    """List the OSPF neighbors for the given router and IP version."""

    if ip_version == 4:
        cmd = "show ip ospf neighbor json"
    else:
        cmd = "show ipv6 ospf neighbor json"

    output = router.vtysh_cmd(cmd)

    if ip_version == 4:
        return [v for n in json.loads(output)["neighbors"].values() for v in n]
    else:
        return json.loads(output)["neighbors"]


def ospf_neighbor_uptime(router, interface, ip_version):
    """Uptime of the neighbor with the given interface name in seconds."""

    for neighbor in ospf_neighbors(router, ip_version):
        if ip_version == 4:
            if not neighbor["ifaceName"].startswith("{}:".format(interface)):
                continue

            return neighbor["upTimeInMsec"] / 1000
        else:
            if neighbor["interfaceName"] != interface:
                continue

            h, m, s = [int(d) for d in neighbor["duration"].split(":")]
            return h * 3600 + m * 60 + s

    raise KeyError(
        "No IPv{} neighbor with interface name {} on {}".format(
            ip_version, interface, router.name
        )
    )


def ospf_routes(router, prefix):
    """List the OSPF routes for the given router and prefix."""

    if ipaddress.ip_interface(prefix).ip.version == 4:
        cmd = "show ip route {} json"
    else:
        cmd = "show ipv6 route {} json"

    output = router.vtysh_cmd(cmd.format(prefix))
    return json.loads(output)[prefix]


def ospf_nexthops(router, prefix, protocol):
    """List the OSPF nexthops for the given prefix."""

    for route in ospf_routes(router, prefix):
        if route["protocol"] != protocol:
            continue

        for nexthop in route["nexthops"]:
            yield nexthop


def ospf_directly_connected_interfaces(router, ip_version):
    """The names of the directly connected interfaces, as discovered
    through the OSPF nexthops.

    """

    if ip_version == 4:
        prefix = "192.0.2.{}/32".format(router.name.strip("r"))
    else:
        prefix = "fe80::/64"

    hops = ospf_nexthops(router, prefix, protocol="connected")
    return sorted([n["interfaceName"] for n in hops if n["directlyConnected"]])


def wait_for_ospf(router, ip_version, neighbors, timeout=60):
    """Wait until the router has the given number of neighbors that are
    fully converged.

    Note that this checks for the exact number of neighbors, so if one neighbor
    is requested and three are converged, the wait continues.

    """

    until = time.monotonic() + timeout

    if ip_version == 4:
        filter = {"converged": "Full"}
    else:
        filter = {"state": "Full"}

    def is_match(neighbor):
        for k, v in filter.items():
            if neighbor[k] != v:
                return False

        return True

    while time.monotonic() < until:
        found = sum(1 for n in ospf_neighbors(router, ip_version) if is_match(n))

        if neighbors == found:
            return

    raise TimeoutError(
        "Waited over {}s for {} neighbors to reach {}".format(
            timeout, neighbors, filter
        )
    )


@pytest.mark.parametrize("ip_version", [4, 6])
def test_interface_up(tgen, ip_version):
    """Verify the initial routing table, before any changes."""

    # Wait for the routers to be ready
    routers = {id: tgen.gears[id] for id in ("r1", "r2", "r3", "r4")}

    for router in routers.values():
        wait_for_ospf(router, ip_version=ip_version, neighbors=3)

    # Verify that the link-local routes are correct
    for router in routers.values():
        connected = ospf_directly_connected_interfaces(router, ip_version)

        if ip_version == 4:
            expected = ["eth1", "eth2", "eth3", "lo"]
        else:
            expected = ["eth1", "eth2", "eth3"]

        assert (
            connected == expected
        ), "Expected all interfaces to be connected on {}".format(router.name)


@pytest.mark.parametrize("ip_version", [4, 6])
def test_interface_down(tgen, ip_version):
    """Verify the routing table after taking interfaces down."""

    # Wait for the routers to be ready
    routers = {id: tgen.gears[id] for id in ("r1", "r2", "r3", "r4")}

    for id, router in routers.items():
        wait_for_ospf(router, ip_version=ip_version, neighbors=3)

    # Keep track of the uptime of the eth3 neighbor
    uptime = ospf_neighbor_uptime(routers["r1"], "eth3", ip_version)
    before = time.monotonic()

    # Take the links between r1 and r2 down
    routers["r1"].cmd_raises("ip link set down dev eth1")
    routers["r1"].cmd_raises("ip link set down dev eth2")

    # Wait for OSPF to converge
    wait_for_ospf(routers["r1"], ip_version=ip_version, neighbors=1)

    # The uptime of the unaffected eth3 neighbor should be monotonic
    new_uptime = ospf_neighbor_uptime(routers["r1"], "eth3", ip_version)
    took = round(time.monotonic() - before, 3)

    # IPv6 has a resolution of 1s, for IPv4 some slack is necesssary.
    if ip_version == 4:
        offset = 0.25
    else:
        offset = 1

    assert (
        new_uptime + offset >= uptime + took
    ), "The eth3 neighbor uptime must not decrease"

    # We should only find eth3 once OSPF has converged
    connected = ospf_directly_connected_interfaces(routers["r1"], ip_version)

    if ip_version == 4:
        expected = ["eth3", "lo"]
    else:
        expected = ["eth3"]

    assert connected == expected, "Expected only eth1 and eth2 to be disconnected"


@pytest.mark.parametrize("ip_version", [4, 6])
def test_interface_flap(tgen, ip_version):
    """Verify the routing table after enabling an interface that was down."""

    # Wait for the routers to be ready
    routers = {id: tgen.gears[id] for id in ("r1", "r2", "r3", "r4")}

    for id, router in routers.items():
        wait_for_ospf(router, ip_version=ip_version, neighbors=3)

    # Keep track of the uptime of the eth3 neighbor
    uptime = ospf_neighbor_uptime(routers["r1"], "eth3", ip_version)
    before = time.monotonic()

    # Take the links between r1 and r2 down
    routers["r1"].cmd_raises("ip link set down dev eth1")
    routers["r2"].cmd_raises("ip link set down dev eth2")

    # Wait for OSPF to converge
    wait_for_ospf(routers["r1"], ip_version=ip_version, neighbors=1)

    # Take the links between r1 and r2 up
    routers["r1"].cmd_raises("ip link set up dev eth1")
    routers["r2"].cmd_raises("ip link set up dev eth2")

    # Wait for OSPF to converge
    wait_for_ospf(routers["r1"], ip_version=ip_version, neighbors=3)

    # The uptime of the unaffected eth3 neighbor should be monotonic
    new_uptime = ospf_neighbor_uptime(routers["r1"], "eth3", ip_version)
    took = round(time.monotonic() - before, 3)

    # IPv6 has a resolution of 1s, for IPv4 some slack is necesssary.
    if ip_version == 4:
        offset = 0.25
    else:
        offset = 1

    assert (
        new_uptime + offset >= uptime + took
    ), "The eth3 neighbor uptime must not decrease"

    # We should find all interfaces again
    connected = ospf_directly_connected_interfaces(routers["r1"], ip_version)

    if ip_version == 4:
        expected = ["eth1", "eth2", "eth3", "lo"]
    else:
        expected = ["eth1", "eth2", "eth3"]

    assert connected == expected, "Expected all interfaces to be connected"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2024 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, get_topogen

pytestmark = [pytest.mark.bgpd]


def setup_module(mod):
    topodef = {"s1": ("r1", "r2"), "s2": ("r2", "r3"), "s3": ("r3", "r4")}
    tgen = Topogen(topodef, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_frr_config(os.path.join(CWD, "{}/frr.conf".format(rname)))

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_dynamic_capability_role():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r2 = tgen.gears["r2"]
    r3 = tgen.gears["r3"]
    r4 = tgen.gears["r4"]

    def _bgp_converge(router):
        output = json.loads(router.vtysh_cmd("show bgp ipv4 unicast json detail"))
        expected = {
            "routes": {
                "10.10.10.40/32": {
                    "paths": [
                        {
                            "extendedIpv6Community": {
                                "string": "LB:65000:5000000000 (40.000 Gbps)",
                            }
                        }
                    ]
                },
                "10.10.10.100/32": {
                    "paths": [
                        {
                            "extendedIpv6Community": {
                                "string": "LB:65000:12500000000 (100.000 Gbps)",
                            }
                        }
                    ]
                },
                "10.10.10.200/32": {
                    "paths": [
                        {
                            "extendedIpv6Community": {
                                "string": "LB:65000:25000000000 (200.000 Gbps)",
                            }
                        }
                    ]
                },
            }
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_converge,
        r2,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert result is None, "r2 (iBGP) should see link bandwidth extended communities"

    test_func = functools.partial(
        _bgp_converge,
        r3,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert (
        result is None
    ), "r3 (eBGP) should see link bandwidth extended communities (including non-transitive)"

    def _bgp_check_non_transitive_extended_communities():
        output = json.loads(r4.vtysh_cmd("show bgp ipv4 unicast json detail"))
        expected = {
            "routes": {
                "10.10.10.40/32": {
                    "paths": [
                        {
                            "extendedIpv6Community": None,
                        }
                    ]
                },
                "10.10.10.100/32": {
                    "paths": [
                        {
                            "extendedIpv6Community": {
                                "string": "LB:65000:12500000000 (100.000 Gbps)",
                            }
                        }
                    ]
                },
                "10.10.10.200/32": {
                    "paths": [
                        {
                            "extendedIpv6Community": {
                                "string": "LB:65000:25000000000 (200.000 Gbps)",
                            }
                        }
                    ]
                },
            }
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(
        _bgp_check_non_transitive_extended_communities,
    )
    _, result = topotest.run_and_expect(test_func, None, count=30, wait=1)
    assert (
        result is None
    ), "r4 (eBGP) should NOT see non-transitive link bandwidth extended communities"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2020 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation, Inc.
# ("NetDEF") in this file.
#


"""OSPF Basic Functionality Automation."""
import os
import sys
import time
import pytest
from copy import deepcopy
from ipaddress import IPv4Address

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

# Import topoJson from lib, to create topology and initial configuration
from lib.common_config import (
    start_topology,
    write_test_header,
    write_test_footer,
    reset_config_on_routers,
    step,
    create_interfaces_cfg,
    retry,
    run_frr_cmd,
)
from lib.topolog import logger
from lib.topojson import build_config_from_json
from lib.topotest import frr_unicode, json_cmp

from lib.ospf import (
    verify_ospf_interface,
)


pytestmark = [pytest.mark.ospfd, pytest.mark.staticd]

# Global variables
topo = None


"""
TOPOOLOGY =
      Please view in a fixed-width font such as Courier.
      +---+  A1       +---+
      +R1 +------------+R2 |
      +-+-+-           +--++
        |  --        --  |
        |    -- A0 --    |
      A0|      ----      |
        |      ----      | A2
        |    --    --    |
        |  --        --  |
      +-+-+-            +-+-+
      +R0 +-------------+R3 |
      +---+     A3     +---+

TESTCASES =
1. OSPF P2MP -Verify state change events on p2mp network.
 """


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """
    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/ospf_p2mp.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Running setup_module() done")


def teardown_module():
    """
    Teardown the pytest environment.

    * `mod`: module name
    """

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


# ##################################
# Test cases start here.
# ##################################


def test_ospf_p2mp_tc1_p0(request):
    """OSPF IFSM -Verify state change events on p2mp network."""
    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    global topo
    step("Bring up the base config as per the topology")
    reset_config_on_routers(tgen)
    step(
        "Verify that OSPF is subscribed to multi cast services "
        "(All SPF, all DR Routers)."
    )
    step("Verify that interface is enabled in ospf.")
    step("Verify that config is successful.")
    dut = "r0"
    input_dict = {
        "r0": {
            "links": {
                "r3": {"ospf": {"mcastMemberOspfAllRouters": True, "ospfEnabled": True}}
            }
        }
    }
    result = verify_ospf_interface(tgen, topo, dut=dut, input_dict=input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Delete the ip address")
    topo1 = {
        "r0": {
            "links": {
                "r3": {
                    "ipv4": topo["routers"]["r0"]["links"]["r3"]["ipv4"],
                    "interface": topo["routers"]["r0"]["links"]["r3"]["interface"],
                    "delete": True,
                }
            }
        }
    }

    result = create_interfaces_cfg(tgen, topo1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Change the ip on the R0 interface")

    topo_modify_change_ip = deepcopy(topo)
    intf_ip = topo_modify_change_ip["routers"]["r0"]["links"]["r3"]["ipv4"]
    topo_modify_change_ip["routers"]["r0"]["links"]["r3"]["ipv4"] = str(
        IPv4Address(frr_unicode(intf_ip.split("/")[0])) + 3
    ) + "/{}".format(intf_ip.split("/")[1])

    build_config_from_json(tgen, topo_modify_change_ip, save_bkup=False)
    step("Verify that interface is enabled in ospf.")
    dut = "r0"
    input_dict = {
        "r0": {
            "links": {
                "r3": {
                    "ospf": {
                        "ipAddress": topo_modify_change_ip["routers"]["r0"]["links"][
                            "r3"
                        ]["ipv4"].split("/")[0],
                        "ipAddressPrefixlen": int(
                            topo_modify_change_ip["routers"]["r0"]["links"]["r3"][
                                "ipv4"
                            ].split("/")[1]
                        ),
                    }
                }
            }
        }
    }
    result = verify_ospf_interface(tgen, topo, dut=dut, input_dict=input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Modify the mask on the R0 interface")
    ip_addr = topo_modify_change_ip["routers"]["r0"]["links"]["r3"]["ipv4"]
    mask = topo_modify_change_ip["routers"]["r0"]["links"]["r3"]["ipv4"]
    step("Delete the ip address")
    topo1 = {
        "r0": {
            "links": {
                "r3": {
                    "ipv4": ip_addr,
                    "interface": topo["routers"]["r0"]["links"]["r3"]["interface"],
                    "delete": True,
                }
            }
        }
    }

    result = create_interfaces_cfg(tgen, topo1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step("Change the ip on the R0 interface")

    topo_modify_change_ip = deepcopy(topo)
    intf_ip = topo_modify_change_ip["routers"]["r0"]["links"]["r3"]["ipv4"]
    topo_modify_change_ip["routers"]["r0"]["links"]["r3"]["ipv4"] = str(
        IPv4Address(frr_unicode(intf_ip.split("/")[0])) + 3
    ) + "/{}".format(int(intf_ip.split("/")[1]) + 1)

    build_config_from_json(tgen, topo_modify_change_ip, save_bkup=False)
    step("Verify that interface is enabled in ospf.")
    dut = "r0"
    input_dict = {
        "r0": {
            "links": {
                "r3": {
                    "ospf": {
                        "ipAddress": topo_modify_change_ip["routers"]["r0"]["links"][
                            "r3"
                        ]["ipv4"].split("/")[0],
                        "ipAddressPrefixlen": int(
                            topo_modify_change_ip["routers"]["r0"]["links"]["r3"][
                                "ipv4"
                            ].split("/")[1]
                        ),
                    }
                }
            }
        }
    }
    result = verify_ospf_interface(tgen, topo, dut=dut, input_dict=input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    topo1 = {
        "r0": {
            "links": {
                "r3": {
                    "ipv4": topo_modify_change_ip["routers"]["r0"]["links"]["r3"][
                        "ipv4"
                    ],
                    "interface": topo_modify_change_ip["routers"]["r0"]["links"]["r3"][
                        "interface"
                    ],
                    "delete": True,
                }
            }
        }
    }

    result = create_interfaces_cfg(tgen, topo1)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    build_config_from_json(tgen, topo, save_bkup=False)

    step("Change the area id on the interface")
    input_dict = {
        "r0": {
            "links": {
                "r3": {
                    "interface": topo["routers"]["r0"]["links"]["r3"]["interface"],
                    "ospf": {"area": "0.0.0.0"},
                    "delete": True,
                }
            }
        }
    }

    result = create_interfaces_cfg(tgen, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    input_dict = {
        "r0": {
            "links": {
                "r3": {
                    "interface": topo["routers"]["r0"]["links"]["r3"]["interface"],
                    "ospf": {"area": "0.0.0.1"},
                }
            }
        }
    }

    result = create_interfaces_cfg(tgen, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)
    step("Verify that interface is enabled in ospf.")
    dut = "r0"
    input_dict = {
        "r0": {"links": {"r3": {"ospf": {"area": "0.0.0.1", "ospfEnabled": True}}}}
    }
    result = verify_ospf_interface(tgen, topo, dut=dut, input_dict=input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    input_dict = {
        "r0": {
            "links": {
                "r3": {
                    "interface": topo["routers"]["r0"]["links"]["r3"]["interface"],
                    "ospf": {"area": "0.0.0.1"},
                    "delete": True,
                }
            }
        }
    }

    result = create_interfaces_cfg(tgen, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    input_dict = {
        "r0": {
            "links": {
                "r3": {
                    "interface": topo["routers"]["r0"]["links"]["r3"]["interface"],
                    "ospf": {"area": "0.0.0.0"},
                }
            }
        }
    }

    result = create_interfaces_cfg(tgen, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    step("Verify if interface is enabled with network type P2MP")
    input_dict = {
        "r0": {
            "links": {
                "r3": {
                    "interface": topo["routers"]["r0"]["links"]["r3"]["interface"],
                    "ospf": {"area": "0.0.0.0", "networkType": "POINTOMULTIPOINT"},
                }
            }
        }
    }
    result = create_interfaces_cfg(tgen, input_dict)
    assert result is True, "Testcase {} :Failed \n Error: {}".format(tc_name, result)

    write_test_footer(tc_name)


def test_ospf_p2mp_tc_delay_reflood(request):
    """OSPF IFSM -Verify "delay-reflood" parameter in p2mp network."""
    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()
    r0 = tgen.gears["r0"]

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    global topo

    step("Verify for interface with network type P2MP that delay-reflood is configured")
    r0.vtysh_multicmd(
        "conf t\ninterface r0-r1-eth0\nip ospf network point-to-multipoint delay-reflood"
    )

    dut = "r0"
    input_dict = {
        "r0": {
            "links": {
                "r1": {
                    "ospf": {
                        "mcastMemberOspfAllRouters": True,
                        "ospfEnabled": True,
                        "networkType": "POINTOMULTIPOINT",
                        "p2mpDelayReflood": True,
                    }
                },
                "r2": {
                    "ospf": {
                        "mcastMemberOspfAllRouters": True,
                        "ospfEnabled": True,
                        "networkType": "POINTOMULTIPOINT",
                        "p2mpDelayReflood": False,
                    }
                },
                "r3": {
                    "ospf": {
                        "mcastMemberOspfAllRouters": True,
                        "ospfEnabled": True,
                        "networkType": "POINTOMULTIPOINT",
                        "p2mpDelayReflood": False,
                    }
                },
            }
        }
    }
    result = verify_ospf_interface(tgen, topo, dut=dut, input_dict=input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)
    delay_reflood_cfg = (
        tgen.net["r0"]
        .cmd(
            'vtysh -c "show running" | grep "^ ip ospf network point-to-multipoint delay-reflood"'
        )
        .rstrip()
    )

    assertmsg = "delay-reflood' configuration applied, but not present in configuration"
    assert (
        delay_reflood_cfg == " ip ospf network point-to-multipoint delay-reflood"
    ), assertmsg

    step("Verify for interface with network type P2MP that delay-reflood is removed")
    r0.vtysh_multicmd(
        "conf t\ninterface r0-r1-eth0\nip ospf network point-to-multipoint"
    )

    input_dict = {
        "r0": {
            "links": {
                "r1": {
                    "ospf": {
                        "mcastMemberOspfAllRouters": True,
                        "ospfEnabled": True,
                        "networkType": "POINTOMULTIPOINT",
                        "p2mpDelayReflood": False,
                    }
                },
                "r2": {
                    "ospf": {
                        "mcastMemberOspfAllRouters": True,
                        "ospfEnabled": True,
                        "networkType": "POINTOMULTIPOINT",
                        "p2mpDelayReflood": False,
                    }
                },
                "r3": {
                    "ospf": {
                        "mcastMemberOspfAllRouters": True,
                        "ospfEnabled": True,
                        "networkType": "POINTOMULTIPOINT",
                        "p2mpDelayReflood": False,
                    }
                },
            }
        }
    }
    result = verify_ospf_interface(tgen, topo, dut=dut, input_dict=input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    delay_reflood_cfg = (
        tgen.net["r0"]
        .cmd(
            'vtysh -c "show running" | grep "^ ip ospf network point-to-multipoint delay-reflood"'
        )
        .rstrip()
    )
    assertmsg = (
        "delay-reflood' configuration removed, but still present in configuration"
    )
    assert (
        delay_reflood_cfg != " ip ospf network point-to-multipoint delay-reflood"
    ), assertmsg

    step(
        "Verify for interface with network type P2MP that delay-reflood is removed with removal of network type"
    )
    r0.vtysh_multicmd(
        "conf t\ninterface r0-r1-eth0\nip ospf network point-to-multipoint delay-reflood"
    )
    r0.vtysh_multicmd(
        "conf t\ninterface r0-r1-eth0\nno ip ospf network point-to-multipoint"
    )
    r0.vtysh_multicmd(
        "conf t\ninterface r0-r1-eth0\nip ospf network point-to-multipoint"
    )

    input_dict = {
        "r0": {
            "links": {
                "r1": {
                    "ospf": {
                        "mcastMemberOspfAllRouters": True,
                        "ospfEnabled": True,
                        "networkType": "POINTOMULTIPOINT",
                        "p2mpDelayReflood": False,
                    }
                },
                "r2": {
                    "ospf": {
                        "mcastMemberOspfAllRouters": True,
                        "ospfEnabled": True,
                        "networkType": "POINTOMULTIPOINT",
                        "p2mpDelayReflood": False,
                    }
                },
                "r3": {
                    "ospf": {
                        "mcastMemberOspfAllRouters": True,
                        "ospfEnabled": True,
                        "networkType": "POINTOMULTIPOINT",
                        "p2mpDelayReflood": False,
                    }
                },
            }
        }
    }
    result = verify_ospf_interface(tgen, topo, dut=dut, input_dict=input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    delay_reflood_cfg = (
        tgen.net["r0"]
        .cmd(
            'vtysh -c "show running" | grep "^ ip ospf network point-to-multipoint delay-reflood"'
        )
        .rstrip()
    )
    assertmsg = (
        "delay-reflood' configuration removed, but still present in configuration"
    )
    assert (
        delay_reflood_cfg != " ip ospf network point-to-multipoint delay-reflood"
    ), assertmsg

    write_test_footer(tc_name)


@retry(retry_timeout=30)
def verify_ospf_json(tgen, dut, input_dict, cmd="show ip ospf database json"):
    del tgen
    show_ospf_json = run_frr_cmd(dut, cmd, isjson=True)
    if not bool(show_ospf_json):
        return "ospf is not running"
    result = json_cmp(show_ospf_json, input_dict)
    return str(result) if result else None


@pytest.mark.parametrize("tgen", [2], indirect=True)
def test_ospf_nbrs(tgen):
    db_full = {
        "areas": {
            "0.0.0.0": {
                "routerLinkStates": [
                    {
                        "lsId": "100.1.1.0",
                        "advertisedRouter": "100.1.1.0",
                        "numOfRouterLinks": 6,
                    },
                    {
                        "lsId": "100.1.1.1",
                        "advertisedRouter": "100.1.1.1",
                        "numOfRouterLinks": 6,
                    },
                    {
                        "lsId": "100.1.1.2",
                        "advertisedRouter": "100.1.1.2",
                        "numOfRouterLinks": 6,
                    },
                    {
                        "lsId": "100.1.1.3",
                        "advertisedRouter": "100.1.1.3",
                        "numOfRouterLinks": 7,
                    },
                ]
            }
        }
    }
    input = [
        [
            "r0",
            "show ip ospf n json",
            {
                "neighbors": {
                    "100.1.1.1": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                    "100.1.1.2": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                    "100.1.1.3": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                }
            },
        ],
        [
            "r1",
            "show ip ospf n json",
            {
                "neighbors": {
                    "100.1.1.0": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                    "100.1.1.2": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                    "100.1.1.3": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                }
            },
        ],
        [
            "r2",
            "show ip ospf n json",
            {
                "neighbors": {
                    "100.1.1.0": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                    "100.1.1.1": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                    "100.1.1.3": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                }
            },
        ],
        [
            "r3",
            "show ip ospf n json",
            {
                "neighbors": {
                    "100.1.1.0": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                    "100.1.1.1": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                    "100.1.1.2": [
                        {
                            "nbrState": "Full/DROther",
                        }
                    ],
                }
            },
        ],
        ["r0", "show ip ospf database json", db_full],
        ["r1", "show ip ospf database json", db_full],
        ["r2", "show ip ospf database json", db_full],
        ["r3", "show ip ospf database json", db_full],
        ["r0", "show ip ospf database json", db_full],
        ["r0", "show ip ospf database router json", {}],
        ["r0", "show ip ospf interface traffic json", {}],
        ["r1", "show ip ospf interface traffic json", {}],
        ["r2", "show ip ospf interface traffic json", {}],
        ["r3", "show ip ospf interface traffic json", {}],
    ]
    for cmd_set in input:
        step("test_ospf: %s - %s" % (cmd_set[0], cmd_set[1]))
        assert (
            verify_ospf_json(tgen, tgen.gears[cmd_set[0]], cmd_set[2], cmd_set[1])
            is None
        )


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC
#
# test_zebra_rib.py
#
# Copyright (c) 2019 by
# Cumulus Networks, Inc
# Donald Sharp
#

"""
test_zebra_rib.py: Test some basic zebra <-> kernel interactions
"""

import os
import re
import sys
from functools import partial
import pytest
import json
import platform

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger
from time import sleep


pytestmark = [pytest.mark.sharpd]
krel = platform.release()


def config_macvlan(tgen, r_str, device, macvlan):
    "Creates specified macvlan interace on physical device"

    if topotest.version_cmp(krel, "5.1") < 0:
        return

    router = tgen.gears[r_str]
    router.run(
        "ip link add {} link {} type macvlan mode bridge".format(macvlan, device)
    )
    router.run("ip link set {} up".format(macvlan))


def setup_module(mod):
    "Sets up the pytest environment"
    # 8 links to 8 switches on r1
    topodef = {"s{}".format(x): ("r1",) for x in range(1, 9)}
    tgen = Topogen(topodef, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()
    for rname, router in router_list.items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_SHARP, os.path.join(CWD, "{}/sharpd.conf".format(rname))
        )

    # Macvlan interface for protodown func test */
    config_macvlan(tgen, "r1", "r1-eth0", "r1-eth0-macvlan")
    # Initialize all routers.
    tgen.start_router()


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()
    tgen.stop_topology()


def test_zebra_kernel_route_vrf():
    "Test kernel routes should be removed after interface changes vrf"
    logger.info("Test kernel routes should be removed after interface changes vrf")
    vrf = "RED"
    tgen = get_topogen()
    r1 = tgen.gears["r1"]

    # Add kernel routes, the interface is initially in default vrf
    r1.run("ip route add 3.5.1.0/24 via 192.168.210.1 dev r1-eth0")
    json_file = "{}/r1/v4_route_1_vrf_before.json".format(CWD)
    expected = json.loads(open(json_file).read())
    test_func = partial(
        topotest.router_json_cmp, r1, "show ip route 3.5.1.0/24 json", expected
    )
    _, result = topotest.run_and_expect(test_func, None, count=5, wait=1)
    assert result is None, '"r1" JSON output mismatches'

    # Change the interface's vrf
    r1.run("ip link add {} type vrf table 1".format(vrf))
    r1.run("ip link set {} up".format(vrf))
    r1.run("ip link set dev r1-eth0 master {}".format(vrf))

    expected = "{}"
    test_func = partial(
        topotest.router_output_cmp, r1, "show ip route 3.5.1.0/24 json", expected
    )
    result, diff = topotest.run_and_expect(test_func, "", count=5, wait=1)
    assertmsg = "{} should not have the kernel route.\n{}".format('"r1"', diff)
    assert result, assertmsg

    # Clean up
    r1.run("ip link set dev r1-eth0 nomaster")
    r1.run("ip link del dev {}".format(vrf))


def test_zebra_kernel_admin_distance():
    "Test some basic kernel routes added that should be accepted"
    logger.info("Test some basic kernel routes that should be accepted")
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip("skipped because of router(s) failure")

    r1 = tgen.gears["r1"]

    # Route with 255/8192 metric

    distance = 255
    metric = 8192

    def makekmetric(dist, metric):
        return (dist << 24) + metric

    r1.run(
        "ip route add 4.5.1.0/24 via 192.168.210.2 dev r1-eth0 metric "
        + str(makekmetric(255, 8192))
    )
    # Route with 1/1 metric
    r1.run(
        "ip route add 4.5.2.0/24 via 192.168.211.2 dev r1-eth1 metric "
        + str(makekmetric(1, 1))
    )
    # Route with 10/1 metric
    r1.run(
        "ip route add 4.5.3.0/24 via 192.168.212.2 dev r1-eth2 metric "
        + str(makekmetric(10, 1))
    )
    # Same route with a 160/1 metric
    r1.run(
        "ip route add 4.5.3.0/24 via 192.168.213.2 dev r1-eth3 metric "
        + str(makekmetric(160, 1))
    )

    # Currently I believe we have a bug here with the same route and different
    # metric.  That needs to be properly resolved.  Making a note for
    # coming back around later and fixing this.
    # tgen.mininet_cli()
    for i in range(1, 3):
        json_file = "{}/r1/v4_route_{}.json".format(CWD, i)
        expected = json.loads(open(json_file).read())

        test_func = partial(
            topotest.router_json_cmp,
            r1,
            "show ip route 4.5.{}.0 json".format(i),
            expected,
        )
        _, result = topotest.run_and_expect(test_func, None, count=20, wait=0.5)
        assertmsg = '"r1" JSON output mismatches'
        assert result is None, assertmsg
    # tgen.mininet_cli()


def test_zebra_kernel_override():
    "Test that a FRR route with a lower admin distance takes over"
    logger.info("Test kernel override with a better admin distance")
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip("skipped because of previous test failure")

    r1 = tgen.gears["r1"]
    r1.vtysh_cmd("conf\nip route 4.5.1.0/24 192.168.216.3")
    json_file = "{}/r1/v4_route_1_static_override.json".format(CWD)
    expected = json.loads(open(json_file).read())

    test_func = partial(
        topotest.router_json_cmp, r1, "show ip route 4.5.1.0 json", expected
    )
    _, result = topotest.run_and_expect(test_func, None, count=20, wait=0.5)
    assert result is None, '"r1" JSON output mismatches'

    logger.info(
        "Test that the removal of the static route allows the kernel to take back over"
    )
    r1.vtysh_cmd("conf\nno ip route 4.5.1.0/24 192.168.216.3")
    json_file = "{}/r1/v4_route_1.json".format(CWD)
    expected = json.loads(open(json_file).read())

    test_func = partial(
        topotest.router_json_cmp, r1, "show ip route 4.5.1.0 json", expected
    )
    _, result = topotest.run_and_expect(test_func, None, count=20, wait=0.5)
    assert result is None, '"r1" JSON output mismatches'


def test_route_map_usage():
    "Test that FRR only reruns over routes associated with the routemap"
    logger.info("Test that FRR runs on selected re's on route-map changes")
    tgen = get_topogen()
    if tgen.routers_have_failure():
        pytest.skip("Skipped because of previous test failure")

    thisDir = os.path.dirname(os.path.realpath(__file__))

    r1 = tgen.gears["r1"]
    # set the delay timer to 1 to improve test coverage (HA)
    r1.vtysh_cmd("conf\nzebra route-map delay-timer 1")
    r1.vtysh_cmd("conf\nroute-map static permit 10\nset src 192.168.215.1")
    r1.vtysh_cmd("conf\naccess-list 5 seq 5 permit 10.0.0.44/32")
    r1.vtysh_cmd("conf\naccess-list 10 seq 5 permit 10.0.1.0/24")
    r1.vtysh_cmd(
        "conf\nroute-map sharp permit 10\nmatch ip address 10\nset src 192.168.214.1"
    )
    r1.vtysh_cmd("conf\nroute-map sharp permit 20\nset src 192.168.213.1")
    r1.vtysh_cmd("conf\nip protocol static route-map static")
    r1.vtysh_cmd("conf\nip protocol sharp route-map sharp")
    sleep(4)
    r1.vtysh_cmd("conf\nip route 10.100.100.100/32 192.168.216.3")
    r1.vtysh_cmd("conf\nip route 10.100.100.101/32 10.0.0.44")
    r1.vtysh_cmd("sharp install route 10.0.0.0 nexthop 192.168.216.3 500")

    def check_initial_routes_installed(router):
        output = json.loads(router.vtysh_cmd("show ip route summ json"))
        expected = {
            "routes": [{"type": "static", "rib": 2}, {"type": "sharp", "rib": 500}]
        }
        return topotest.json_cmp(output, expected)

    test_func = partial(check_initial_routes_installed, r1)
    _, result = topotest.run_and_expect(test_func, None, count=40, wait=1)

    static_rmapfile = "%s/r1/static_rmap.ref" % (thisDir)
    expected = open(static_rmapfile).read().rstrip()
    expected = ("\n".join(expected.splitlines()) + "\n").rstrip()
    logger.info(
        "Does the show route-map static command run the correct number of times"
    )

    def check_static_map_correct_runs():
        actual = r1.vtysh_cmd("show route-map static")
        actual = re.sub(r"\([0-9].* milli", "(X milli", actual)
        actual = ("\n".join(actual.splitlines()) + "\n").rstrip()
        return topotest.get_textdiff(
            actual,
            expected,
            title1="Actual Route-map output",
            title2="Expected Route-map output",
        )

    ok, result = topotest.run_and_expect(
        check_static_map_correct_runs, "", count=10, wait=1
    )
    assert ok, result

    sharp_rmapfile = "%s/r1/sharp_rmap.ref" % (thisDir)
    expected = open(sharp_rmapfile).read().rstrip()
    expected = ("\n".join(expected.splitlines()) + "\n").rstrip()
    logger.info("Does the show route-map sharp command run the correct number of times")

    def check_sharp_map_correct_runs():
        actual = r1.vtysh_cmd("show route-map sharp")
        actual = re.sub(r"\([0-9].* milli", "(X milli", actual)
        actual = ("\n".join(actual.splitlines()) + "\n").rstrip()
        return topotest.get_textdiff(
            actual,
            expected,
            title1="Actual Route-map output",
            title2="Expected Route-map output",
        )

    ok, result = topotest.run_and_expect(
        check_sharp_map_correct_runs, "", count=10, wait=1
    )
    assert ok, result

    logger.info(
        "Add a extension to the static route-map to see the static route go away"
        " and test that the routes installed are correct"
    )

    r1.vtysh_cmd("conf\nroute-map sharp deny 5\nmatch ip address 5")
    # we are only checking the kernel here as that this will give us the implied
    # testing of both the route-map and staticd withdrawing the route
    # let's spot check that the routes were installed correctly
    # in the kernel
    sharp_ipfile = "%s/r1/iproute.ref" % (thisDir)
    expected = open(sharp_ipfile).read().rstrip()
    expected = ("\n".join(expected.splitlines()) + "\n").rstrip()

    def check_routes_installed():
        actual = r1.run("ip route show")
        actual = ("\n".join(actual.splitlines()) + "\n").rstrip()
        actual = re.sub(r" nhid [0-9][0-9]", "", actual)
        actual = re.sub(r" proto sharp", " proto XXXX", actual)
        actual = re.sub(r" proto static", " proto XXXX", actual)
        actual = re.sub(r" proto 194", " proto XXXX", actual)
        actual = re.sub(r" proto 196", " proto XXXX", actual)
        actual = re.sub(r" proto kernel", " proto XXXX", actual)
        actual = re.sub(r" proto 2", " proto XXXX", actual)
        # Some platforms have double spaces?  Why??????
        actual = re.sub(r"  proto XXXX  ", " proto XXXX ", actual)
        actual = re.sub(r"  metric", " metric", actual)
        actual = re.sub(r" link  ", " link ", actual)
        return topotest.get_textdiff(
            actual,
            expected,
            title1="Actual ip route show",
            title2="Expected ip route show",
        )

    ok, result = topotest.run_and_expect(check_routes_installed, "", count=5, wait=1)
    assert ok, result


def test_protodown():
    "Run protodown basic functionality test and report results."
    pdown = False
    count = 0
    tgen = get_topogen()
    if topotest.version_cmp(krel, "5.1") < 0:
        tgen.errors = "kernel 5.1 needed for protodown tests"
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]

    # Set interface protodown on
    r1.vtysh_cmd("sharp interface r1-eth0-macvlan protodown")

    # Timeout to wait for dplane to handle it
    while count < 10:
        count += 1
        output = r1.vtysh_cmd("show interface r1-eth0-macvlan")
        if re.search(r"protodown reasons:.*sharp", output):
            pdown = True
            break
        sleep(1)

    assert pdown is True, "Interface r1-eth0-macvlan not set protodown"

    # Set interface protodown off
    r1.vtysh_cmd("no sharp interface r1-eth0-macvlan protodown")

    # Timeout to wait for dplane to handle it
    while count < 10:
        count += 1
        output = r1.vtysh_cmd("show interface r1-eth0-macvlan")
        if not re.search(r"protodown reasons:.*sharp", output):
            pdown = False
            break
        sleep(1)

    assert pdown is False, "Interface r1-eth0-macvlan not set protodown off"


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
